@@ -67,6 +67,10 @@
  * PWS 1996/12/10
  */
 
+#ifdef HAVE_STDIO_H
+#include <stdio.h>
+#endif
+
 #include "zsh.mdh"
 #include "input.pro"
 
@@ -219,11 +223,16 @@ static int
 inputline(void)
 {
     char *ingetcline, *ingetcpmptl = NULL, *ingetcpmptr = NULL;
+    int context = ZLCON_LINE_START;
 
     /* If reading code interactively, work out the prompts. */
     if (interact && isset(SHINSTDIN)) {
-	if (!isfirstln)
+	if (!isfirstln) {
 	    ingetcpmptl = prompt2;
+	    if (rprompt2)
+		ingetcpmptr = rprompt2;
+	    context = ZLCON_LINE_CONT;
+	}
 	else {
 	    ingetcpmptl = prompt;
 	    if (rprompt)
@@ -262,8 +271,11 @@ inputline(void)
 	 * typeahead when the terminal settings are altered.
 	 *                     pws 1998/03/12
 	 */
-	ingetcline = (char *)zleread(ingetcpmptl, ingetcpmptr,
-				     ZLRF_HISTORY|ZLRF_NOSETTY);
+	int flags = ZLRF_HISTORY|ZLRF_NOSETTY;
+	if (isset(IGNOREEOF))
+	    flags |= ZLRF_IGNOREEOF;
+	ingetcline = (char *)zleread(ingetcpmptl, ingetcpmptr, flags,
+				     context);
 	histdone |= HISTFLAG_SETTY;
     }
     if (!ingetcline) {
